program read_dom;

{$mode objfpc}{$H+}

uses
  Classes, SysUtils, laz2_xmlread, laz2_DOM;

const
  fn = 'data.xml';

function GetNodeValue(ANode: TDOMNode): String;
var
  child: TDOMNode;
begin
  Result := '';
  child := ANode.FirstChild;
  if Assigned(child) and (child.NodeName = '#text') then
    Result := child.NodeValue;
end;

procedure ReadSharedStrings(ANode: TDOMNode; AList: TStrings);
var
  valuenode: TDOMNode;
  childnode: TDOMNode;
  nodename: String;
  txt: String;
begin
  while Assigned(ANode) do begin
    if ANode.NodeName = 'si' then begin
      txt := '';
      valuenode := ANode.FirstChild;
      while valuenode <> nil do begin
        nodename := valuenode.NodeName;
        if nodename = 't' then
        begin
          txt := GetNodeValue(valuenode);
          AList.Add(txt);
        end;
        valuenode := valuenode.NextSibling;
      end;
    end;
    ANode := ANode.NextSibling;
  end;
end;

procedure ReadXMLStream(out ADoc: TXMLDocument; AStream: TStream);
var
  parser: TDOMParser;
  src: TXMLInputSource;
begin
  parser := TDOMParser.Create;
  try
    parser.Options.PreserveWhiteSpace := true;
    src := TXMLInputSource.Create(AStream);
    try
      parser.Parse(src, ADoc);
    finally
      src.Free;
    end;
  finally
    parser.Free;
  end;
end;

procedure ReadFile(AFileName: String; AList: TStrings);
var
  stream: TMemoryStream;
  doc: TXMLDocument;
  t0: TDateTime;
begin
  stream := TMemoryStream.Create;
  try
    stream.LoadFromFile(AFileName);
    stream.Position := 0;
    t0 := Now();
    ReadXMLStream(Doc, stream);
    try
      ReadSharedStrings(Doc.DocumentElement.FindNode('si'), AList);
    finally
      FreeAndNil(Doc);
    end;
    WriteLn(FormatDateTime('nn:ss.zzz', Now - t0));
  finally
    stream.Free;
  end;
end;

var
  FSharedStrings: TStrings;
  i: Integer;

begin
  FSharedStrings := TStringList.Create;
  try
    ReadFile(fn, FSharedStrings);
    WriteLn(FSharedStrings.Count, ' <si> nodes found. The first 10 nodes are:');
    for i:=0 to 9 do
      WriteLn('  ''', FSharedStrings[i], '''');
  finally
    FSharedStrings.Free;
  end;

  WriteLn('Press ENTER to close.');
  ReadLn;
end.
