program read_sax;

{$mode objfpc}{$H+}

uses
  Classes, SysUtils, Strings, sax, sax_xml;

const
  fn = 'data.xml';

function GetSAXText(ch: PSAXChar; ALength: Integer): String;
var
  ws: widestring;
begin
  if (ALength > 0) then begin
    SetLength(ws, ALength);
    StrLCopy(PWideChar(ws), PWideChar(ch), ALength);
    Result := UTF8Encode(ws);
  end else
    Result := '';
end;

function GetSAXAttrValue(Atts: TSAXAttributes; ALocalName: String): String;
var
  i: Integer;
begin
  Result := '';
  if Atts = nil then
    exit;
  for i:=0 to Atts.Length-1 do
    if Atts.LocalNames[i] = ALocalName then
    begin
      Result := Atts.Values[i];
      exit;
    end;
end;


{ TXLSXSharedStringsReader }

type
  TXLSXSharedStringsReader = class(TSAXXMLReader)
  private
    FSharedStrings: TStrings;
    FCurrentText: String;
    siNode: Boolean;
    tNode: Boolean;
    preserveSpace: Boolean;
    procedure CharactersHandler(Sender: TObject; const ch: PSAXChar;
      AStart, ALength: Integer);
    procedure EndElementHandler(Sender: TObject;
      const NamespaceURI, LocalName, QName: SAXString);
    procedure StartElementHandler(Sender: TObject;
      const NamespaceURI, LocalName, QName: SAXString; Atts: TSAXAttributes);
  public
    constructor Create(ASharedStrings: TStrings);
  end;


constructor TXLSXSharedStringsReader.Create(ASharedStrings: TStrings);
begin
  inherited Create;
  FSharedStrings := ASharedStrings;
  OnCharacters := @CharactersHandler;
  OnStartElement := @StartElementHandler;
  OnEndElement := @EndElementHandler;
end;

procedure TXLSXSharedStringsReader.CharactersHandler(Sender: TObject;
  const ch: PSAXChar; AStart, ALength: Integer);
var
  s: String;
begin
  if tNode then begin
    s := GetSAXText(ch, ALength);
    FCurrentText := FCurrentText + s;
  end;
end;

procedure TXLSXSharedStringsReader.StartElementHandler(Sender: TObject;
  const NamespaceURI, LocalName, QName: SAXString; Atts: TSAXAttributes);
var
  i: Integer;
begin
  case LocalName of
    'si':
      begin
        siNode := true;
        FCurrentText := '';
      end;
    't':
      begin
        tNode := true;
        preserveSpace := GetSAXAttrValue(Atts, 'xml:space') = 'preserve';
      end;
  end;
end;

procedure TXLSXSharedStringsReader.EndElementHandler(Sender: TObject;
  const NamespaceURI, LocalName, QName: SAXString);
begin
  case LocalName of
    'si':
      if siNode then
      begin
        siNode := false;
        FSharedStrings.Add(FCurrentText);
        FCurrentText := '';
      end;
    't':
      if tNode then
      begin
        if preserveSpace and (FCurrentText = '') then
          FCurrentText := ' ';
        preserveSpace := false;
      end;
  end;
end;

procedure ReadFile(AFileName: String; AList: TStrings);
var
  stream: TMemoryStream;
  inputSource: TSAXInputSource;
  reader: TSAXXMLReader;
  t0: TDateTime;
begin
  stream := TMemoryStream.Create;
  try
    stream.LoadFromFile(AFileName);
    stream.Position := 0;
    inputsource := TSAXInputSource.Create(stream);
    t0 := now();
    reader := TXLSXSharedStringsReader.Create(AList);
    try
      reader.Parse(inputSource);
      WriteLn(FormatDateTime('nn:ss.zzz', now - t0), ' ms');
    finally
      reader.Free;
      inputSource.Free;
    end;
  finally
    stream.Free;
  end;
end;

var
  FSharedStrings: TStrings;
  i: Integer;

begin
  FSharedStrings := TStringList.Create;
  try
    ReadFile(fn, FSharedStrings);
    WriteLn(FSharedStrings.Count, ' <si> nodes found. The first 10 nodes are:');
    for i:=0 to 9 do
      WriteLn('  ''', FSharedStrings[i], '''');
  finally
    FSharedStrings.Free;
  end;

  WriteLn('Press ENTER to close.');
  ReadLn;
end.

